/***********************************************************************
 *  Intelligent file output routines
 *  (c) 1997 by C.Schroeter (clausi@chemie.fu-berlin.de)
 *
 *  adapted from an idea from harald boehme and martin loewis
 *
 *  the IF routines provide a differential write to files
 *  so files will not be really touched until they really change
 *  this avoids full remakes with minor changes to the ddl files
 *
 ***********************************************************************/


#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <assert.h>
#include "iwrite.h"

#include <stdarg.h>

extern int debug;

/* ---------- ---------------------------------------------------*/
IFILE *IFopen(char *name) {

 IFILE *ifd;
 
 if(( ifd = (IFILE *)malloc( sizeof(IFILE))) == NULL ) 
  return NULL;
 strcpy(ifd->name,name);
 ifd->pos=0;
 ifd->backup=1;

 if(( ifd->handle = fopen(ifd->name,"r")) == NULL ) {
   strcat( strcpy(ifd->new_name,ifd->name),".new");
   if(( ifd->handle = fopen(ifd->name,"w")) == NULL ) {
     free(ifd);
     return NULL;
   }
   ifd->changed = 1;
   ifd->backup=0;
 }
 ifd->changed = 0; 

} 

/* ---------- ---------------------------------------------------*/

int IFwrite(IFILE *ifd, char *str,int len ) {
 static char buffer[4096];
 int i;

 assert( len <= strlen(str) );
 if(ifd->changed) {
   /* write to the real file */
   i = fwrite( str, 1, len, ifd->handle);
   if( i!= len ) return -1;
   ifd->pos += len;
   return len;
 } /* else */
 assert(len < sizeof(buffer));
 i = fread( buffer, 1, len, ifd->handle );
 if( i==len && strncmp(str, buffer, len) == 0 ) {
   ifd->pos += len;
   return len;
 }
 switch_files(ifd);
 IFwrite(ifd,str,len);
}

static void switch_files(IFILE *ifd) {

  int i;
  long l;
  static char buffer[4096];
  FILE *old_handle = ifd->handle;

  /* open a new file */
  strcat( strcpy(ifd->new_name,ifd->name),".new");
  ifd->handle = fopen(ifd->new_name,"w");
  if( ifd->handle == NULL ) {
    fprintf(stderr,"IFwrite: cannot open %s for writing\n",ifd->new_name);
    exit (-1);
  }
  ifd->changed = 1;
  ifd->backup  = 1;

  if(debug) fprintf(stderr,"IFwrite: '%s' changed, backup=%d \n", ifd->name, ifd->backup );

  l = fseek(old_handle, 0L, SEEK_SET);  
  if( l != 0 ) {
    fprintf(stderr,"IFwrite: problem with seek\n");
    exit (-1);
  }
  while( ifd->pos > sizeof(buffer) ) {
    i = fread(buffer, 1, sizeof(buffer), old_handle);
    if( i != sizeof(buffer) ) {
      fprintf(stderr,"IFwrite: problem with read\n");
      exit (-1);
    }
    i = fwrite(buffer, 1, sizeof(buffer), ifd->handle);
    if( i != sizeof(buffer) ) {
      fprintf(stderr,"IFwrite: problem with write\n");
      exit (-1);
    }
    ifd->pos -= sizeof(buffer);
  }    
  i = fread( buffer, 1, (unsigned) ifd->pos, old_handle);
    if( i != ifd->pos ) {
      fprintf(stderr,"IFwrite: problem with read\n");
      exit (-1);
    }
  i = fwrite( buffer, 1, (unsigned) ifd->pos, ifd->handle);  
    if( i != ifd->pos ) {
      fprintf(stderr,"IFwrite: problem with write\n");
      exit (-1);
    }
  i = fclose(old_handle);
  ifd->pos = 0;
}


/* ---------- ---------------------------------------------------*/

void IFclose(IFILE *ifd) {
  int ret;
  if( ifd->handle && ifd->handle != stdout && ifd->handle != stderr) {
    if( ! ifd->changed ) {
      fread(&ret,1,1,ifd->handle);
      if( !feof(ifd->handle) ) 
	switch_files(ifd);
    }
  }
  ret = fclose(ifd->handle);
  if(ret) {
    fprintf(stderr,"IFclose: Error on closing %s\n",ifd->name);
    perror("fclose");
  }
  if( ifd->changed ) {
    if( ifd->backup ) {
      char *buf = malloc(strlen(ifd->name)+5);
      sprintf(buf,"%s.bak",ifd->name);
      rename(ifd->name,buf);
      free(buf);
      ifd->backup=0;
    }
    ret = rename( ifd->new_name, ifd->name );
    if( ret != 0) {
      fprintf(stderr,"IFclose: Error on renaming %s to %s ",ifd->new_name,ifd->name);
      perror("fclose");      
    }
    free(ifd);
  }
}

/* ---------- ---------------------------------------------------*/

void IFputs(char *s, IFILE *fd) {
  IFwrite(fd,s,strlen(s));
}
/* ---------- ---------------------------------------------------*/

void IFputc(char c,IFILE *fd) {
  IFwrite(fd,&c,1);
}
/* ---------- ---------------------------------------------------*/

int IFprintf(IFILE *fd,char *format, ...) {
static char buffer[4096];
int len;
va_list ap;
va_start(ap,format);

if( strlen(format) > sizeof(buffer) ) {
  fprintf(stderr,"IFprintf: sorry buffer too large");
  exit (-1);
}

vsprintf(buffer, format, ap);
va_end(ap);
len = IFwrite(fd,buffer,strlen(buffer));
return len;
}
