module "rmgr":

/*
 * RMGR Kernel Ressource Manager (c) 1996 C.Schroeter 
 *      (clausi@chemie.fu-berlin.de)
 *
 */

/@
 The rmgr provides a ressource database feature (similar to the Xressource)
 to the kernel. Each ressource can have its own access permissions if accessed
 from the user interface.
 
 If a ressource is accessed a callback can be invoked so actions can be invoked
 depending on a ressource's value. Callbacks cannot be specified 
 from the user interface.

 Kernel ressources are stored in a global ressource database that can be 
 accessed from user space using the rmanip program.

 The rmgr can be configured using the Kernel ressource database. 

 
@/

driver "rmgr" [major=29] 
{
  
  include Sys.ddl

  /* this includes the other ddl's from ./ddlsrc */
  include types.ddl
  include init.ddl
  include install.ddl
  include lookup.ddl
  include destroy.ddl
  include delete.ddl
  include info.ddl
  include strsave.ddl
  include hash.ddl
  include perm.ddl
  include callback.ddl
  include lock.ddl
  include errors.ddl

  /* the core method */
  core {
  init /{ 
	  RCS_entry *np,v;
          extern void TestCallback();
          printk("Ressource Manager initializing\n");
          RCS_init();
  }/
  cleanup /{
	printk("Ressource Manager unloading\n");
	RCS_destroy();	
  }/
  }
  open /{
  }/
  close /{ 
     /*RCS_info();*/ 
  }/
  /* 
   * This are the definitions for the user interface
   * note that each routine name begins with rmgr_
   *
   */
  ioctl { /@
	  Gets the type field information from a ressource entry
	  <name> is the ressource name and <len> its length

          The type can be one of
             RCS_STRING	   a string
	     RCS_INT       an integer
             RCS_IPTR      a pointer to an integer

          If the ressource cannot be found in the database a -1 is
	  returned and errno is set to -ENOTFOUND
          @/
	  func GetType(char *name, int len) /{
		char *buf;
		RCS_entry *np;
                DBGin("GetType");
		if( (buf = (char *) %<driver.name>_GetFromFS((void *) name,len)) == NULL ){
                        DBGout();return -EIO;
                }
		np = RCS_lookup(buf); kfree(buf);
		if( np == NULL ){ 
			Error(-ENOTFOUND);
			DBGout(); return -1;
		}
		DBGout(); return np->type;
	  }/
          /@
	  Changes the user-level access modes of the ressource <name>
          to mode.
          @/
	  func Chmod(char *name, int len, int mode) /{
		  char *buf;
	          char *buf2;
		  RCS_entry entry;
		  RCS_entry *np;
                  DBGin("SetInt");
		  if( (buf = (char *) %<driver.name>_GetFromFS((void *) name,len)) == NULL ){
                        DBGout();return -EIO;
                  }
	          DBGprint(DBG_DATA,("name=%s mode=0x%x",buf,mode));
		  if( (np = RCS_lookup(buf)) != NULL ){ 
		     RCS_lock(np);
                     if( ! RCS_CheckPerm(np,RCS_PERM_WRITE)){
			DBGprint(DBG_DATA,("write permission denied"));
			Error(-ENOPERM); kfree(buf);
			DBGout(); return -1;
		     }				
		     np->mflags = mode;
		     RCS_unlock(np);
                  }
		  kfree(buf);
		  DBGout(); return np->mflags;
	  }/
          /@
	  Set a string type ressource to a new value. If a ressource cannot 
          be found in the kernel database a new one will be created.

	  If the ressource exists but has a different type errno is set to
	  -ENOTYPE
          @/
	  func SetString(char *name, int len,char *sbuf, int slen) /{
		  char *buf;
	          char *buf2;
		  RCS_entry entry;
		  RCS_entry *np;
                  DBGin("SetString");
		  if( (buf = (char *) %<driver.name>_GetFromFS((void *) name,len)) == NULL ){
                        DBGout();return -EIO;
                  }
		  if( (buf2 = (char *) %<driver.name>_GetFromFS((void *) sbuf,slen)) == NULL ){
                        DBGout();return -EIO;
                  }
	          DBGprint(DBG_DATA,("name=%s string=%s",buf,buf2));
		  if( (np = RCS_lookup(buf)) != NULL ){ 
		     if( np->type != RCS_STRING ){
			DBGprint(DBG_DATA,("type conflict"));
			Error(-ENOTYPE);kfree(buf);kfree(buf2);
			DBGout(); return -1;
		     }
                     if( ! RCS_CheckPerm(np,RCS_PERM_WRITE)){
			DBGprint(DBG_DATA,("write permission denied"));
			Error(-ENOPERM);kfree(buf);kfree(buf2);
			DBGout(); return -1;
		     }
                     RCS_lock(np);
		     entry.type = RCS_STRING;
		     entry.uval.sval = buf2;
                     entry.mflags = np->mflags;
		     RCS_install(buf,&entry);
		     RCS_unlock(np);
                  } else {
		     entry.type = RCS_STRING;
		     entry.uval.sval = buf2;
                     entry.mflags = O_RD | O_WR;
		     RCS_install(buf,&entry);
   		     DBGprint(DBG_DATA,("wrote %s as %s",buf,entry.uval.sval));
                  }
		  kfree(buf);
                  kfree(buf2);
		  DBGout(); return 1;
	  }/
	/@
	Get the value of a string type variable from the kernel databese
        in sbuf. If the ressource is unexistent a -1 is returned and
	errno is set to -ENOTFOUND
        @/
          func GetString(char *name, int len,char *sbuf) /{
		  char *buf;
		  RCS_entry entry;
		  RCS_entry *np;
                  int slen=0;
                  DBGin("GetString");
		  if( (buf = (char *) %<driver.name>_GetFromFS((void *) name,len)) == NULL ){
                        DBGout();return -EIO;
                  }
	          DBGprint(DBG_DATA,("Lookup String name=%s",buf));
		  np = RCS_lookup(buf);
		  if( np == NULL ){
			DBGprint(DBG_DATA,("entry %s not found",buf));
			Error(-ENOTFOUND);kfree(buf);
			DBGout(); return -1;
                  }
                  kfree(buf);
		  RCS_lock(np);
                  if( ! RCS_CheckPerm(np,RCS_PERM_READ)){
			DBGprint(DBG_DATA,("read permission denied"));
			Error(-ENOPERM);
			DBGout(); return -1;
	          }				
		  if(( slen = strlen(np->uval.sval)) > 0){
                        %<driver.name>_PutToFS(sbuf,np->uval.sval,slen);
                  }
		  RCS_unlock(np);
		  DBGout(); return slen;
	  }/
	  /@
           Assigns function type ressource to a target pointer
           both ressources have to exist in the kernel database.

           This function is intended to provide a generic function 
           interface. So certain functionalities of a driver can be
           modified throug the ressource manager.

           Functions can only be set if the consuming driver
           is already loaded in the kernel.
	  @/
          func AssignFunc(char *to, int tolen, char *from, int frlen) /{ 
		char *buf1,*buf2;
                RCS_entry *np1,*np2;

	  	DBGin("AssignFunc");
	        if( (buf1 = (char *) 
                     %<driver.name>_GetFromFS((void *) to,tolen)) == NULL ){
                     DBGout();return -EIO;
                }
	        if( (buf2 = (char *) 
                     %<driver.name>_GetFromFS((void *) from,frlen)) == NULL ){
                     kfree(buf1);
                     DBGout();return -EIO;
                }
	        DBGprint(DBG_DATA,("to=%s from=%s",buf1,buf2));
                if((np2 = RCS_lookup(buf2)) == NULL ){
		    kfree(buf1);kfree(buf2);
		    DBGout();return -ENOPERM; 
		}		
                if((np1 = RCS_lookup(buf1)) == NULL ){
		    kfree(buf1);kfree(buf2);
		    DBGout();return -ENOPERM; 
		}		
		if( (np1->type != RCS_FPTR) || (np2->type != RCS_FPTR)){
		    kfree(buf1);kfree(buf2);
		    DBGout();return -ENOTYPE; 
		}
		np1->uval.fptr = np2->uval.fptr;
                if( np1->callback != NULL ){
	          DBGprint(DBG_DATA,("calling callback"));
	          (*np1->callback)(RCS_CHANGE_EVENT,np1);
                }
	        DBGprint(DBG_DATA,("%s changed to %lx ",buf1,np1->uval.fptr));
                kfree(buf1);kfree(buf2);
		DBGout();
          }/

          /@
	  Set a integer type ressource to a new value. If a ressource cannot 
          be found in the kernel database a new one will be created.
          @/
	  func SetInt(char *name, int len, int aint) /{
		  char *buf;
		  RCS_entry entry;
		  RCS_entry *np;
		  extern int umask;

                  DBGin("SetInt");
		  if( (buf = (char *) %<driver.name>_GetFromFS((void *) name,len)) == NULL ){
                        DBGout();return -EIO;
                  }
	          DBGprint(DBG_DATA,("name=%s int=%d",buf,aint));

		  if( (np = RCS_lookup(buf)) != NULL ){ 
		     if( np->type != RCS_INT && np->type != RCS_IPTR ){
			DBGprint(DBG_DATA,("type conflict %d",np->type));
			Error(-ENOTYPE);kfree(buf);
			DBGout(); return -1;
		     }
                     if( ! RCS_CheckPerm(np,RCS_PERM_WRITE)){
			DBGprint(DBG_DATA,("write permission denied"));
			DBGout(); return -1;
		     }
		     RCS_lock(np);
                     if( np->type == RCS_INT )
   		           np->uval.ival = aint;
                     else
			   *np->uval.iptr = aint;				
		     RCS_install(buf,np);
		     RCS_unlock(np);
                  } else {
		     entry.type = RCS_INT;
		     entry.uval.ival = aint;
                     entry.mflags = umask;
		     RCS_install(buf,&entry);
  		     DBGprint(DBG_DATA,("wrote %s as %d",buf,entry.uval.ival));
                  }
		  kfree(buf);
		  DBGout();return 1;
	  }/
	/@
	Get the value of a integer type variable from the kernel databese
        in sbuf. If the ressource is unexistent a -1 is returned.
        If the ressource is a RCS_IPTR type the value of the referenced
        pointer will be replaced instead of its address.
        @/
	  func GetInt(char *name,int len) /{
		char *buf;
		RCS_entry *np;
                DBGin("GetInt");
		if( (buf = (char *) %<driver.name>_GetFromFS((void *) name,len)) == NULL ){
                        DBGout();return -EIO;
                }
		np = RCS_lookup(buf);
		if( np == NULL ){
			DBGprint(DBG_DATA,("entry not found"));
			Error(-ENOTFOUND);kfree(buf);
			DBGout(); return -1;
                }
                DBGprint(DBG_DATA,("name=%s int=%d",buf,np->uval.ival));
		kfree(buf);
		RCS_lock(np);
                if( ! RCS_CheckPerm(np,RCS_PERM_READ)){
			DBGprint(DBG_DATA,("read permission denied"));
			DBGout(); return -1;
		}				
                RCS_unlock(np);
		DBGout(); 
		if(np->type == RCS_INT)
			return np->uval.ival;
		else
			return *np->uval.iptr;
	  }/
          /@
	  Deletes a ressource from the kernel ressource database.
	  If the ressource is nonexistent a -1 will be returned.
          @/
	  func Delete(char *name,int len) /{
		char *buf;
                RCS_entry *np;
                DBGin("Delete");
		if( (buf = (char *) %<driver.name>_GetFromFS((void *) name,len)) == NULL ){
                        DBGout();return -EIO;
                }
                if(( np = RCS_lookup(buf)) != NULL ){
                 if( ! RCS_CheckPerm(np,RCS_PERM_WRITE) && !(np->mflags & PERM_STICKY) ){
			DBGprint(DBG_DATA,("delete permission denied"));
			Error(-ENOPERM);
			DBGout(); return -1;
		 }				
	         RCS_delete(buf);
                }
                kfree(buf);
		DBGout(); return 1;		
	  }/
	  func Info() /{
		DBGin("Info");
		RCS_info();
		DBGout();
	  }/
	}

}


